from memobj.property import *

from .properties import *


class HashNode(MemoryObject):
    left: "HashNode" = DereffedPointer(0x0, "HashNode")
    parent: "HashNode" = DereffedPointer(0x8, "HashNode")
    right: "HashNode" = DereffedPointer(0x10, "HashNode")
    is_leaf: bool = Bool(0x19)
    hash: int = Signed4(0x20)
    node_data: "Type" = DereffedPointer(0x28, "Type")


class Type(MemoryObject):
    name: str = CppString(0x38)
    hash: int = Signed4(0x58)
    size: int = Signed4(0x60)
    name_2: str = CppString(0x68)
    is_pointer: bool = Bool(0x88)
    if_ref: bool = Bool(0x89)
    property_list: "PropertyList" = DereffedPointer(0x90, "PropertyList")

    def get_bases(self) -> list["PropertyList"]:
        if (fields := self.property_list) is None:
            return []

        bases = []
        current_base = fields
        while (base_type := current_base.base_class_list) is not None:
            bases.append(base_type)
            current_base = base_type

        return bases


class PropertyList(MemoryObject):
    is_singleton: bool = Bool(0x9)
    offset: int = Signed4(0x10)
    base_class_list: "PropertyList" = DereffedPointer(0x18, "PropertyList")
    type: "Type" = DereffedPointer(0x20, "Type")
    pointer_version: "Type" = DereffedPointer(0x30, "Type")
    properties: list["Property"] = SharedVector(0x58, object_type="Property")
    functions: list["Function"] = SharedVector(0x70, object_type="Function")
    name: str = CppString(0xB8)


class Property(MemoryObject):
    list: "PropertyList" = DereffedPointer(0x38, "PropertyList")
    container: "Container" = DereffedPointer(0x40, "Container")
    index: int = Signed4(0x50)
    name: str = DereffedPointer(0x58, NullTerminatedString(None, search_size=100))
    name_hash: int = Signed4(0x60)
    full_hash: int = Signed4(0x64)
    offset: int = Signed4(0x68)
    type: "Type" = DereffedPointer(0x70, "Type")
    flags: int = Signed4(0x80)
    note: str = CppString(0x88)
    ps_info: str = CppString(0x90)
    enum_options = PropertyEnumOptions(0x98)


class Function(MemoryObject):
    list: "PropertyList" = DereffedPointer(0x30, "PropertyList")
    name: str = CppString(0x38)
    details: "FunctionDetails" = DereffedPointer(0x58, "FunctionDetails")


class Container(MemoryObject):
    vtable: int = Unsigned8(0x0)
    name: str = ContainerName(None)
    is_dynamic: bool = ContainerIsDynamic(None)


class FunctionDetails(MemoryObject):
    called_function: int = Unsigned8(0x30)
    something: int = Unsigned4(0x3C)



from memobj import WindowsProcess

from .memory import HashNode

HASH_CALL_PATTERN = rb"\xE8....\x48\x3B\x18\x74\x12"


def _get_root_node(process: WindowsProcess) -> HashNode:
    # for now this pattern is unique in WizardGraphicalClient.exe
    # note that there is an ambiguous match in the bug reporter,
    # which is why we restrict the scan here
    hash_call_addr = process.scan_memory(HASH_CALL_PATTERN, module="WizardGraphicalClient.exe")[0]

    # E8 [B2 43 00 00]
    call_offset = process.read_formatted(hash_call_addr + 1, "i")

    # 5 is the length of the call instruction
    call_addr = hash_call_addr + call_offset + 5

    # 48 8B 05 [BF 0A F7 01]
    hash_tree_offset = process.read_formatted(call_addr + 53, "i")

    # 50 is start of the lea instruction and 7 is the length of it
    hash_tree_addr = call_addr + 50 + hash_tree_offset + 7

    pointer = process.read_formatted(hash_tree_addr, "Q")
    address = process.read_formatted(pointer, "Q")

    return HashNode(address=address, process=process)


def _get_children_nodes(node: HashNode, nodes: set):
    nodes.add(node)

    if node.is_leaf is False:
        if left_node := node.left:
            if left_node not in nodes:
                _get_children_nodes(left_node, nodes)
        if right_node := node.right:
            if right_node not in nodes:
                _get_children_nodes(right_node, nodes)

    return nodes


def _read_all_nodes(root_node: HashNode):
    nodes = set()

    first_node = root_node.parent
    return _get_children_nodes(first_node, nodes)


def get_hash_nodes(process: WindowsProcess) -> set[HashNode]:
    root_node = _get_root_node(process)
    return _read_all_nodes(root_node)


def get_type_tree() -> dict[str, HashNode]:
    process = WindowsProcess.from_name("WizardGraphicalClient.exe")

    nodes = get_hash_nodes(process)

    hash_map = {}

    for node in nodes:
        if node.is_leaf:
            continue

        hash_map[node.node_data.name] = node

    return hash_map



from typing import Any

from iced_x86 import Code, Decoder, Instruction, Register
from memobj import MemoryObject, MemoryProperty


class CppString(MemoryProperty):
    def __init__(self, offset: int | None, encoding: str = "utf-8", sso_size: int = 16):
        super().__init__(offset)
        self.encoding = encoding
        self.sso_size = sso_size

    def from_memory(self) -> Any:
        length = self.process.read_formatted(
            self.offset_address + 16,
            self.pointer_format_string,
        )
        capacity = self.process.read_formatted(
            self.offset_address + 16 + 8,
            self.pointer_format_string,
        )
        
        if capacity >= self.sso_size:
            address = self.read_formatted_from_offset(self.pointer_format_string)
        else:
            address = self.offset_address

        return self.process.read_memory(address, length).decode(self.encoding)

    def to_memory(self, value: Any):
        raise NotImplementedError()

    def memory_size(self) -> int:
        return 32


# TODO rework this into a MemoryObject
class SharedVector(MemoryProperty):
    def __init__(
            self,
            offset: int | None,
            max_size: int = 500,
            object_type: type[MemoryObject] | str | None = None,
    ):
        super().__init__(offset)
        self.max_size = max_size
        self.object_type = object_type

    def from_memory(self) -> Any:
        head = self.read_formatted_from_offset(self.pointer_format_string)
        tail = self.memory_object.memobj_process.read_formatted(
            self.memory_object.base_address + self.offset + 8,
            self.pointer_format_string
        )

        size = tail - head
        element_number = size // 16

        # less than 0 on dealloc
        if size <= 0:
            return []

        if element_number > self.max_size:
            raise ValueError(f"Size was {element_number} and the max was {self.max_size}")

        element_data = self.memory_object.memobj_process.read_memory(head, size)

        pointers = []
        data_position = 0
        for _ in range(element_number):
            pointers.append(int.from_bytes(element_data[data_position:data_position+8], "little", signed=False))
            # 8 byte pointer, 8 byte ref data*
            data_position += 16

        if self.object_type is None:
            return pointers

        if isinstance(self.object_type, str):
            typed_object_type = MemoryObject.__memory_object_instances__.get(self.object_type)

            if typed_object_type is None:
                raise ValueError(f"No MemoryObject type named {self.object_type}")

            self.object_type = typed_object_type

        objects = []
        for pointer in pointers:
            objects.append(self.object_type(
                address=pointer,
                process=self.memory_object.memobj_process,
            ))

        return objects

    def to_memory(self, value: Any):
        pass

    def memory_size(self) -> int:
        return self.pointer_size * 3


class PropertyEnumOptions(MemoryProperty):
    def read_cpp_string(self, address: int, *, sso_size: int = 16, encoding: str = "utf-8"):
        length = self.memory_object.memobj_process.read_formatted(
            address + 16,
            self.pointer_format_string,
        )

        if length >= sso_size:
            address = self.memory_object.memobj_process.read_formatted(address, self.pointer_format_string)
        else:
            address = address

        try:
            return self.memory_object.memobj_process.read_memory(address, length).decode(encoding)
        except UnicodeDecodeError:
            return ""

    def from_memory(self) -> Any:
        start = self.read_formatted_from_offset(self.pointer_format_string)

        if start == 0:
            return None

        end = self.memory_object.memobj_process.read_formatted(
            self.memory_object.base_address + 0xA0,
            self.pointer_format_string
        )

        total_size = end - start

        current = start
        enum_opts = {}
        for entry in range(total_size // 0x48):
            name = self.read_cpp_string(current + 0x28)

            # TODO: int variants are always returned as string
            if string_value := self.read_cpp_string(current):
                enum_opts[name] = string_value
            else:
                enum_opts[name] = self.memory_object.memobj_process.read_formatted(
                    current + 0x20, "I"
                )

            current += 0x48

        return enum_opts

    def to_memory(self, value: Any):
        raise NotImplementedError()

    def memory_size(self) -> int:
        return self.pointer_size * 2


class ContainerName(MemoryProperty):
    def from_memory(self) -> Any:
        # noinspection PyUnresolvedReferences
        vtable = self.memory_object.vtable
        lea_func_addr = self.memory_object.memobj_process.read_formatted(vtable + 0x8, "q")
        name_offset = self.memory_object.memobj_process.read_formatted(lea_func_addr + 3, "i")
        # I need to read a null terminated string, but I can't use my MemoryProperty
        # this is an oversight
        string_bytes = self.memory_object.memobj_process.read_memory(
            lea_func_addr + 7 + name_offset,
            20,
        )

        end = string_bytes.find(b"\x00")

        if end == 0:
            return ""

        if end == -1:
            raise ValueError("No null end")

        return string_bytes[:end].decode()

    def to_memory(self, value: Any):
        raise NotImplementedError()

    # TODO: same as below
    def memory_size(self) -> int:
        return 0


class ContainerIsDynamic(MemoryProperty):
    def from_memory(self) -> Any:
        # noinspection PyUnresolvedReferences
        vtable = self.memory_object.vtable
        get_dynamic_func_addr = self.memory_object.memobj_process.read_formatted(
            vtable + 0x20, self.pointer_format_string
        )

        get_dynamic_bytes = self.memory_object.memobj_process.read_memory(get_dynamic_func_addr, 3)

        decoder = Decoder(64, get_dynamic_bytes)

        xor_al_al = Instruction.create_reg_reg(
            Code.XOR_R8_RM8,
            Register.AL,
            Register.AL
        )
        mov_al_1 = Instruction.create_reg_i32(
            Code.MOV_R8_IMM8,
            Register.AL,
            1
        )

        for instruction in decoder:
            if instruction == xor_al_al:
                return False
            elif instruction == mov_al_1:
                return True
            # Note: ret should never enter here
            else:
                raise RuntimeError(f"Invalid dynamic container instruction: {instruction=}")

    def to_memory(self, value: Any):
        raise NotImplementedError()

    # TODO: this should be a Pointer
    def memory_size(self) -> int:
        return 8



import json
from pathlib import Path


from wiztype.memory import HashNode, Type, Property


class TypeDumper:
    def __init__(self, type_tree: dict[str, HashNode]):
        self.type_tree = type_tree

    def dump(self, output_file: str | Path):
        output = ""
        for formatted_class in self.class_loop(self.type_tree):
            output += formatted_class

        self.output(output_file, output)

    def class_loop(self, type_tree):
        for name, node in type_tree.items():
            data = node.node_data

            formatted_properties = []

            if property_list := data.property_list:
                # print(f"{hex(property_list.base_address)=}")
                # exit()

                for property_ in property_list.properties:
                    formatted_enum_options = []
                    if (enum_options := property_.enum_options) is not None:
                        for enum_option_name, enum_option_value in enum_options.items():
                            formatted_enum_options.append(
                                self.format_enum_option(
                                    enum_option_name, enum_option_value
                                )
                            )

                    property_name, property_info = self.get_property_info(
                        property_
                    )

                    formatted_properties.append(
                        self.format_property(
                            property_name, property_info, formatted_enum_options
                        )
                    )

            base_names, class_hash = self.get_class_info(data)
            formatted_class = self.format_class(
                name, base_names, class_hash, formatted_properties
            )

            yield formatted_class

    @staticmethod
    def output(output_file, output):
        with open(output_file, "w+") as fp:
            fp.write(output)

    @staticmethod
    def get_class_info(node_data: Type):
        bases = node_data.get_bases()
        # & 0xFFFF_FFFF makes it unsigned
        class_hash = node_data.hash & 0xFFFF_FFFF
        return [base.name for base in bases], class_hash

    @staticmethod
    def get_property_info(property_: Property):
        property_name = property_.name
        property_type = property_.type

        container = property_.container
        property_list = property_.list

        property_info = {
            "type": property_type.name,
            "id": property_.index,
            "offset": property_.offset,
            "flags": property_.flags,
            "container": container.name,
            "dynamic": container.is_dynamic,
            "singleton": property_list.is_singleton,
            "pointer": property_type.is_pointer,
            # & 0xFFFF_FFFF makes the value unsigned
            "hash": property_.full_hash & 0xFFFF_FFFF,
        }

        return property_name, property_info

    def format_enum_option(self, name: str, value: int):
        raise NotImplemented()

    def format_property(
        self, name: str, info: dict[str, str], enum_options: list[str]
    ):
        raise NotImplemented()

    def format_class(
        self, name: str, base_names: list[str], class_hash: int, properties: list[str]
    ):
        raise NotImplemented()


class JsonTypeDumperV1(TypeDumper):
    @staticmethod
    def output(output_file, output, indent: int | None = None):
        with open(output_file, "w+") as fp:
            json.dump(output, fp, indent=indent)

    def dump(self, output_file: str | Path, *, indent: int | None = None):
        output = {}
        for formatted_class in self.class_loop(self.type_tree):
            output.update(formatted_class)

        self.output(output_file, output, indent)

    def format_enum_option(self, name: str, value: str):
        return {name: value}

    def format_property(
        self, name: str, info: dict[str, str], enum_options: list[dict]
    ):
        res = {name: info}

        if enum_options:
            options = {}

            for enum_dict in enum_options:
                options.update(enum_dict)

            # noinspection PyTypeChecker
            res[name]["enum_options"] = options

        return res

    def format_class(
        self, name: str, base_names: list[str], class_hash: int, properties: dict
    ):
        props = {}

        for prop in properties:
            props.update(prop)

        return {name: {"bases": base_names, "hash": class_hash, "properties": props}}


class JsonTypeDumperV2(JsonTypeDumperV1):
    version = 2

    def dump(self, output_file: str | Path, *, indent: int | None = None):
        output = {
            "version": self.version,
            "classes": {},
        }
        for formatted_class in self.class_loop(self.type_tree):
            output["classes"].update(formatted_class)

        self.output(output_file, output, indent)

    def format_enum_option(self, name: str, value: int | str):
        try:
            return {name: int(value) & 0xFFFF_FFFF}
        except ValueError:
            return {name: value}

    def format_property(
        self, name: str, info: dict[str, str], enum_options: list[dict]
    ):
        res = {name: info}

        if enum_options:
            options = {}

            for enum_dict in enum_options:
                options.update(enum_dict)

            # noinspection PyTypeChecker
            res[name]["enum_options"] = options

        return res

    def format_class(
        self, name: str, base_names: list[str], class_hash: int, properties: dict
    ):
        props = {}

        for prop in properties:
            props.update(prop)

        return {
            str(class_hash): {
                "name": name,
                "bases": base_names,
                "hash": class_hash,
                "properties": props,
            }
        }


if __name__ == "__main__":
    from wiztype.type_tree import get_type_tree

    tree = get_type_tree()
    dumper = JsonTypeDumperV2(tree)

    dumper.dump("output.json", indent=None)



from pathlib import Path

import click
from memobj import WindowsProcess

import wiztype


@click.command()
@click.argument(
    "outfile",
    type=click.Path(exists=False, dir_okay=False, path_type=Path),
    default="use revision"
)
@click.option("--version", type=int, default=2, show_default=True)
@click.option("--indent", type=int, default=None)
def main(outfile: Path, version: int, indent: int | None):
    match version:
        case 1:
            dump_type = wiztype.JsonTypeDumperV1
        case 2:
            dump_type = wiztype.JsonTypeDumperV2
        case _:
            click.echo(f"{version} is not a supported version")
            exit(1)

    process = WindowsProcess.from_name("WizardGraphicalClient.exe")
    wiz_bin = process.executable_path.parent
    revision_file = wiz_bin / "revision.dat"

    if not revision_file.exists():
        raise FileNotFoundError(f"revision.dat not found in {wiz_bin}")

    revision = revision_file.read_text().strip()

    if str(outfile) == "use revision" or outfile == b"use revision":
        outfile = Path(revision.replace(".", "_") + ".json")

    click.echo(f"dumping types for revision {revision} to {outfile}")

    tree = wiztype.get_type_tree()
    dumper = dump_type(tree)

    dumper.dump(outfile, indent=indent)


if __name__ == "__main__":
    main()
